@@ -13,3 +13,8 @@
 
 #undef ENABLE_HTTP
 #undef  ORBIT_PURIFY
+
+#undef HAVE_SOCKADDR_SA_LEN
+#undef HAVE_INET_PTON
+#undef HAVE_GETADDRINFO
+#undef HAVE_GETNAMEINFO
