@@ -2,7 +2,6 @@
 #include <config.h>
 #include <ctype.h>
 #include <string.h>
-#include <popt.h>
 
 #include <orbit/orbit.h>
 
@@ -38,6 +37,9 @@ static gboolean     orbit_use_http_iors      = FALSE;
 static char        *orbit_ipsock             = NULL;
 static char        *orbit_ipname             = NULL;
 static char        *orbit_debug_options      = NULL;
+static char        *orbit_naming_ref         = NULL;
+static GSList      *orbit_initref_list       = NULL; 
+static GSList      *orbit_use_corbaloc       = FALSE;
 
 void
 ORBit_ORB_start_servers (CORBA_ORB orb)
@@ -243,6 +245,8 @@ shutdown_orb (void)
 
 	if (!(orb = _ORBit_orb))
 		return;
+
+	init_level = 1; /* clobber it */
 	
 	CORBA_exception_init (&ev);
 
@@ -252,6 +256,70 @@ shutdown_orb (void)
 	CORBA_exception_free (&ev);
 }
 
+static
+gboolean
+ORBit_initial_reference_protected_id (gchar* id)
+{
+        return (!strncmp (id, "RootPOA", strlen("RootPOA")) ||
+                !strncmp (id, "POACurrent", strlen("POACurrent")));
+}
+  
+static void 
+ORBit_initial_references_by_user (CORBA_ORB          orb, 
+				  gchar             *naming_ref,
+				  GSList            *initref_list,
+				  CORBA_Environment *ev)
+{
+	GSList *l;
+	CORBA_Object objref;
+
+	if (ev->_major != CORBA_NO_EXCEPTION)
+		return;
+
+	if (naming_ref) {
+		objref = CORBA_ORB_string_to_object (orb, naming_ref, ev);
+		
+		/* FIXME, should abort if invalid option, don't forget
+		 * to free resources allocated by ORB */ 
+		if (ev->_major != CORBA_NO_EXCEPTION) {
+			g_warning ("Option ORBNamingIOR has invalid object reference: %s", 
+				   naming_ref);
+			CORBA_exception_free (ev);
+		} else {
+			/* FIXME, test type of object for
+			 * IDL:omg.org/CosNaming/NamingContext using _is_a()
+			 * operation */ 
+			ORBit_set_initial_reference (orb, "NameService", objref);
+		}
+	}
+
+	for (l = initref_list; l; l = l->next) {
+		ORBit_OptionKeyValue *tuple = l->data;
+
+		g_assert (tuple != NULL);
+		g_assert (tuple->key   != (gchar*)NULL);
+		g_assert (tuple->value != (gchar*)NULL);
+			
+		objref = CORBA_ORB_string_to_object (orb, tuple->value, ev);
+
+		/* FIXME, should abort if invalid option,
+		 * don't forget to free resources allocated by
+		 * ORB */ 
+		if (ev->_major != CORBA_NO_EXCEPTION) {
+			g_warning ("Option ORBInitRef has invalid object reference: %s=%s",  
+				   tuple->key, tuple->value);
+			CORBA_exception_free (ev);
+		}  else if (ORBit_initial_reference_protected_id(tuple->key)) {
+                        g_warning ("Option ORBInitRef permission denied: %s=%s",
+
+                                   tuple->key, tuple->value);
+                } else {
+                        ORBit_set_initial_reference (orb, tuple->key, objref);
+                }
+	}
+
+}
+
 CORBA_ORB
 CORBA_ORB_init (int *argc, char **argv,
 		CORBA_ORBid orb_identifier,
@@ -272,11 +340,14 @@ CORBA_ORB_init (int *argc, char **argv,
 	/* the allocation code uses the bottom bit of any pointer */
 	g_assert (ORBIT_ALIGNOF_CORBA_DOUBLE > 2);
 
-	threaded = (orb_identifier &&
-		    strstr (orb_identifier, "orbit-local-mt-orb") != NULL);
+	if (orb_identifier &&
+	    strstr (orb_identifier, "orbit-local-non-threaded-orb") != NULL)
+		threaded = FALSE;
+	else
+		threaded = TRUE;
 
 	ORBit_option_parse (argc, argv, orbit_supported_options);
-
+	
 #ifdef G_ENABLE_DEBUG
 	ORBit_setup_debug_flags ();
 
@@ -286,6 +357,7 @@ CORBA_ORB_init (int *argc, char **argv,
 	}
 #endif /* G_ENABLE_DEBUG */
 
+
 	giop_recv_set_limit (orbit_initial_recv_limit);
 	giop_init (threaded,
 		   orbit_use_ipv4 || orbit_use_ipv6 ||
@@ -308,6 +380,13 @@ CORBA_ORB_init (int *argc, char **argv,
 
 	retval->adaptors = g_ptr_array_new ();
 	ORBit_init_internals (retval, ev);
+	/* FIXME, handle exceptions */ 
+
+	ORBit_initial_references_by_user (retval, 
+					  orbit_naming_ref,
+					  orbit_initref_list,
+					  ev);
+	/* FIXME, handle exceptions */ 
 
 	return ORBit_RootObject_duplicate (retval);
 }
@@ -331,6 +410,11 @@ CORBA_ORB_object_to_string (CORBA_ORB          orb,
 		return NULL;
 	}
 
+	/* FIXME, do nice integration */ 
+	if (orbit_use_corbaloc) {
+		return ORBit_object_to_corbaloc (obj, ev);
+	}
+	
 	buf = giop_send_buffer_use (orb->default_giop_version);
 
 	g_assert (buf->num_used == 1);
@@ -384,7 +468,12 @@ CORBA_ORB_string_to_object (CORBA_ORB          orb,
 	guchar              *tmpbuf;
 	int                  i;
 
-	if (strncmp (string, "IOR:", 4)) {
+	if (strncmp (string, "IOR:", strlen("IOR:"))            &&
+	    strncmp (string, "corbaloc:", strlen ("corbaloc:")) &&
+	    strncmp (string, "iiop:", strlen ("iiop:"))         &&
+	    strncmp (string, "iiops:", strlen ("iiops:"))       &&
+	    strncmp (string, "ssliop:", strlen ("ssliop:"))     &&
+	    strncmp (string, "uiop:", strlen ("uiop:"))) {
 
 #if defined ENABLE_HTTP
 		if (orbit_use_http_iors &&
@@ -407,41 +496,46 @@ CORBA_ORB_string_to_object (CORBA_ORB          orb,
 		}
 	}
 
-	string += 4;
-	len = strlen (string);
-	while (len > 0 && !g_ascii_isxdigit (string [len - 1]))
-		len--;
-
-	if (len % 2) {
-		if (ior)
-			g_free (ior);
-
-		return CORBA_OBJECT_NIL;
-	}
-
-	tmpbuf = g_alloca (len / 2);
-
-	for (i = 0; i < len; i += 2)
-		tmpbuf [i/2] = (g_ascii_xdigit_value (string [i]) << 4) |
-					g_ascii_xdigit_value (string [i + 1]);
-
-	buf = giop_recv_buffer_use_encaps (tmpbuf, len / 2);
-
-	if (ORBit_demarshal_object (&retval, buf, orb)) {
-		CORBA_exception_set_system (
+	if (!strncmp (string, "IOR:", strlen ("IOR:")))
+	{
+		string += 4;
+		len = strlen (string);
+		while (len > 0 && !g_ascii_isxdigit (string [len - 1]))
+			len--;
+		
+		if (len % 2) {
+			if (ior)
+				g_free (ior);
+			
+			return CORBA_OBJECT_NIL;
+		}
+		
+		tmpbuf = g_alloca (len / 2);
+		
+		for (i = 0; i < len; i += 2)
+			tmpbuf [i/2] = (g_ascii_xdigit_value (string [i]) << 4) |
+				g_ascii_xdigit_value (string [i + 1]);
+		
+		buf = giop_recv_buffer_use_encaps (tmpbuf, len / 2);
+		
+		if (ORBit_demarshal_object (&retval, buf, orb)) {
+			CORBA_exception_set_system (
 				ev,
 				ex_CORBA_MARSHAL,
 				CORBA_COMPLETED_NO);
-
-		retval = CORBA_OBJECT_NIL;
-	}
-
-	giop_recv_buffer_unuse (buf);
-
-	if (ior)
-		g_free (ior);
-
-	return retval;
+			
+			retval = CORBA_OBJECT_NIL;
+		}
+		
+		giop_recv_buffer_unuse (buf);
+		
+		if (ior)
+			g_free (ior);
+		
+		return retval;
+	} else {
+		return ORBit_object_by_corbaloc (orb, string, ev);
+	} 
 }
 
 void
@@ -502,6 +596,33 @@ CORBA_ORB_get_service_information (CORBA_ORB                  orb,
 				   CORBA_ServiceInformation **service_information,
 				   CORBA_Environment         *ev)
 {
+	/* FIXME: 
+         * see http://mail.gnome.org/archives/orbit-list/2003-May/msg00093.html
+
+	 * Assigning NULL to parameter service_information is not
+	 * compliant to CORBA spec. This operation is part of pseudo
+	 * interface and must react like operation of true remote
+	 * interface. The question is what value it should point to in
+	 * case CORBA_FALSE is returned to caller.
+
+	 4.2.2 Getting Service Information
+	 4.2.2.1 get_service_information
+
+	 boolean get_service_information (in ServiceType service_type;
+	                                  out ServiceInformation service_information;);
+
+         * The get_service_information operation is used to obtain
+         * information about CORBA facilities and services that are
+         * supported by this ORB. The service type for which
+         * information is being requested is passed in as the in
+         * parameter service_type, the values defined by constants in
+         * the CORBA module. If service information is available for
+         * that type, that is returned in the out parameter
+         * service_information, and the operation returns the value
+         * TRUE. If no information for the requested services type is
+         * available, the operation returns FALSE (i.e., the service
+         * is not supported by this ORB).
+	 */	
 	*service_information = NULL;
 
 	return CORBA_FALSE;
@@ -556,6 +677,10 @@ CORBA_ORB_list_initial_services (CORBA_ORB          orb,
 	return retval;
 }
 
+/** The InvalidName exception is raised at @ev when
+ *  ORB.resolve_initial_references is passed an @identifier for which
+ *  there is no initial reference.
+ */
 CORBA_Object
 CORBA_ORB_resolve_initial_references (CORBA_ORB          orb,
 				      const CORBA_char  *identifier,
@@ -563,12 +688,22 @@ CORBA_ORB_resolve_initial_references (CORBA_ORB          orb,
 {
 	CORBA_Object objref;
 
-	if (!orb->initial_refs)
+	/* FIXME, verify identifier and raise exception for invalid
+	 * service names, valid names might be: NameService, RootPOA,
+	 * SecurityCurrent, PolicyCurrent, etc. */
+ 
+	if (!orb->initial_refs ||
+	    !(objref = g_hash_table_lookup (orb->initial_refs, identifier)))
 		return CORBA_OBJECT_NIL;
-
-	objref = g_hash_table_lookup (orb->initial_refs, identifier);
-
+	
 	return ORBit_RootObject_duplicate (objref);
+
+ raise_invalid_name:
+	CORBA_exception_set (ev,
+			     CORBA_USER_EXCEPTION,
+			     ex_CORBA_ORB_InvalidName,
+			     NULL);
+	return CORBA_OBJECT_NIL;
 }
 
 static CORBA_TypeCode
@@ -1125,6 +1260,8 @@ ORBit_set_initial_reference (CORBA_ORB  orb,
 		orb->initial_refs = g_hash_table_new (g_str_hash, g_str_equal);
 
 	if (g_hash_table_lookup (orb->initial_refs, identifier))
+               /* FIXME, release object before removing from hash-table,
+                   Frank 2003/06/22 */
 		g_hash_table_remove (orb->initial_refs, identifier);
 
 	g_hash_table_insert (orb->initial_refs,
@@ -1159,7 +1296,8 @@ const ORBit_option orbit_supported_options[] = {
 	{ "ORBid",              ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
 	{ "ORBImplRepoIOR",     ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
 	{ "ORBIfaceRepoIOR",    ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
-	{ "ORBNamingIOR",       ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
+	{ "ORBNamingIOR",       ORBIT_OPTION_STRING,  &orbit_naming_ref},
+
 	{ "ORBRootPOAIOR",      ORBIT_OPTION_STRING,  NULL }, /* FIXME: huh?          */
  	{ "ORBIIOPIPName",      ORBIT_OPTION_STRING,  &orbit_ipname },
  	{ "ORBIIOPIPSock",      ORBIT_OPTION_STRING,  &orbit_ipsock },
@@ -1177,5 +1315,7 @@ const ORBit_option orbit_supported_options[] = {
 	/* warning: this option is a security risk */
 	{ "ORBSimpleUIDs",      ORBIT_OPTION_BOOLEAN, &orbit_use_genuid_simple },
 	{ "ORBDebugFlags",      ORBIT_OPTION_STRING,  &orbit_debug_options },
+	{ "ORBInitRef",         ORBIT_OPTION_KEY_VALUE,  &orbit_initref_list},
+	{ "ORBCorbaloc",        ORBIT_OPTION_BOOLEAN, &orbit_use_corbaloc},
 	{ NULL,                 0,                    NULL },
 };
